#!/usr/bin/env python
#fileencoding: utf-8
#Author: Liu DongMiao <liudongmiao@gmail.com>
#Created  : Fri 07 Jan 2011 10:59:21 AM CST

UPNP_MCAST_ADDR = '239.255.255.250'
UPNP_BCAST_ADDR = '255.255.255.255'
PORT = 1900
SIZE = 1536

from urlparse import urlparse
import time
import socket
import httplib
from xml.dom.minidom import parseString
import sys
from system import set_name_value, get_name_value
from info import report_upnp_info

version = sys.version.split()[0]
if version[0] == '3':
    raise SystemExit('please run under python 2.x')

def _parse_xml(xmlstr):
    """
    Wrapper function of minidom.parseString, which first remove invalid characters and then parse.
    Currently only '0x00' is removed, thanks to Corega and their stupid router CG-WLR300GNE. ;-(
    """
    return parseString(xmlstr.translate(None, '\x00'))

class UPnPException(Exception):
    pass

class UPnP():

    netloc = None
    connection = None
    msearch = 'M-SEARCH * HTTP/1.1\r\n' \
              'HOST: 239.255.255.250:1900\r\n' \
              'MAN: "ssdp:discover"\r\n' \
              'MX: %s\r\n' \
              'ST: %s\r\n' \
              'USER-AGENT: Linux/2.6 UPnP/1.1 python/%s\r\n' \
              '\r\n'

    debug = False
    InternetGateway = 'urn:schemas-upnp-org:device:InternetGatewayDevice:1'
    WANIPConnection = 'urn:schemas-upnp-org:service:WANIPConnection:1'
    WANPPPConnection = 'urn:schemas-upnp-org:service:WANPPPConnection:1'
    UPnPError = 'urn:schemas-upnp-org:control-1-0'

    def __init__(self):
        self.time = 0
        self.timeout = 0
        self.controlurl = None
        pass

    def discover(self, timeout=5, num=1):
        for addr in (UPNP_MCAST_ADDR, UPNP_BCAST_ADDR):
            devs = self._do_discover(addr, timeout, num)
            if devs: return devs

    def _do_discover(self, addr, timeout=5, num=1):
        self.time = int(time.time())
        self.timeout = timeout
        delay = timeout < 5 and timeout or 5
        sudp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        sudp.settimeout(delay)

        msearch = self.msearch % (delay, self.InternetGateway, version)

        if addr == UPNP_BCAST_ADDR:
            sudp.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sudp.sendto(msearch, ('<broadcast>', PORT))
        else:
            serverinfo = socket.getaddrinfo(addr, PORT, socket.AF_UNSPEC, socket.SOCK_DGRAM)
            while serverinfo:
                try:
                    sudp.sendto(msearch, serverinfo.pop(0)[4])
                except socket.error:
                    continue

        devices = []
        while len(devices) < num:
            try:
                result = sudp.recv(SIZE, 0).split('\r\n')
            except socket.timeout:
                break

            if self.debug:
                print result

            st = None
            location = None
            while result:
                line = result.pop(0).strip()
                if line[:2].lower() == 'st':
                    st = line[3:].strip()
                elif line[:8].lower() == 'location':
                    location = line[9:].strip()
                if st == self.InternetGateway and location:
                    devices.append(location)
                    break

        sudp.close()
        return devices

    def descript(self, location):
        try:
            xml = _parse_xml(self._retrive(location))
        except:
            raise UPnPException('Description Error')

        for service in xml.getElementsByTagName('serviceType'):
            servicetype = service.firstChild.data
            if servicetype in (self.WANIPConnection, self.WANPPPConnection):
                controlurl = service.parentNode.getElementsByTagName('controlURL').pop(0).firstChild.data
                if self._get_status_info(controlurl, servicetype):
                    self.controlurl = controlurl
                    break

        xml.unlink()

        if not self.controlurl:
            raise UPnPException('Description without valid WANConnection')

    def _get_status_info(self, controlurl, servicetype):
        ins = []
        outs = {'NewConnectionStatus': '',
                # 'NewLastConnectionError': '',
                # 'NewUptime': '',
                }
        try:
            self.control('GetStatusInfo', ins, outs, controlurl=controlurl, servicetype=servicetype)
        except UPnPException:
            return None
        if outs['NewConnectionStatus'].lower() == 'connected':
            return True
        return None

    def _retrive(self, location, method='GET', headers=None, body=None):
        ''' Get data, without catching exception '''
        timeout = 10
        if self.timeout:
            timeout = self.timeout + self.time - int(time.time())
            if timeout <= 0:
                return ''

        result = urlparse(location)
        if result.scheme and result.netloc != self.netloc:
            if self.connection:
                self.connection.close()
                self.connection = None
            self.netloc = result.netloc

        if self.connection is None:
            self.connection = httplib.HTTPConnection(self.netloc, timeout=timeout)

        if result.query:
            selector = '%s?%s' % (result.path, result.query)
        else:
            selector = result.path
        del result

        self.connection.putrequest(method, selector)
        self.connection.putheader('User-Agent', 'Linux/2.6 UPnP/1.1 Python/%s' % version),

        if headers is not None:
            for header in headers:
                self.connection.putheader(*header)

        if method == 'POST':
            body.seek(0, 2)
            self.connection.putheader('Content-Length', body.tell())
            self.connection.endheaders()
            body.seek(0, 0)
            self.connection.send(body.read())
            if self.debug:
                print
                print '+' * 80, 'send start'
                print '%s %s' % (method, selector)
                print body.getvalue()
                print '+' * 80, 'send end'
            body.close()
        else:
            self.connection.endheaders()
        response = self.connection.getresponse()

        if self.debug:
            print response.status, response.reason
            print response.getheaders()

        try:
            data = '%s' % response.read()
        except:
            data = ''

        connection = response.getheader('Connection', '').lower()
        response.close()

        if connection != 'keep-alive':
            self.connection.close()
            self.connection = None

        if self.debug:
            print '-' * 80, 'receive start'
            print data
            print '-' * 80, 'receive end'
            print

        return data

    def control(self, action, ins=None, outs=None, controlurl=None, servicetype=None):
        from StringIO import StringIO
        if controlurl is None:
            controlurl = self.controlurl
        if servicetype is not None:
            self.servicetype = servicetype
        headers = [('SOAPAction', '"%s#%s"' % (self.servicetype, action)),
                   ('Content-Type', 'text/xml; charset="utf-8"')]
        body = StringIO()

        body.write('<?xml version="1.0"?>')
        body.write('<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"' \
                   ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">' \
                   '<s:Body><u:%s xmlns:u="%s">' % (action, self.servicetype))

        while ins:
            name, value = ins.pop(0)
            body.write('<%s>%s</%s>' % (name, value, name))
            continue
            if value:
                body.write('<%s>%s</%s>' % (name, value, name))
            else:
                body.write('<%s/>' % name)

        body.write('</u:%s></s:Body></s:Envelope>' % action)

        try:
            xml = _parse_xml(self._retrive(controlurl, method='POST', headers=headers, body=body))
        except:
            raise UPnPException('Control %s Error' % action)

        result = None
        responses = xml.getElementsByTagNameNS(self.servicetype, '%sResponse' % action)
        # for possible namespace error
        if not responses:
            body = xml.getElementsByTagNameNS('http://schemas.xmlsoap.org/soap/envelope/', 'Body')
            if body:
                for child in body[0].childNodes:
                    if child.nodeType != child.ELEMENT_NODE:
                        continue
                    if child.localName != '%sResponse' % action:
                        continue
                    responses.append(child)

        if not responses:
            errorCode = xml.getElementsByTagNameNS(self.UPnPError, 'errorCode')
            errorDescription = xml.getElementsByTagNameNS(self.UPnPError, 'errorDescription')
            if errorCode and errorDescription:
                outs['errorCode'] = errorCode.pop(0).firstChild.data
                outs['errorDescription'] = errorDescription.pop(0).firstChild.data
                result = False
            xml.unlink()
            return result

        response = responses.pop(0)
        for name in outs.keys():
            values = response.getElementsByTagName(name)
            if values:
                value = values.pop(0).firstChild
                if hasattr(value, 'toxml'):
                    outs[name] = value.toxml()
                else:
                    outs[name] = None

        xml.unlink()
        return True

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None

def _getspecificportmapping(upnp, remote, port, protocol):
    action = 'GetSpecificPortMappingEntry'
    ins = [('NewRemoteHost', remote),
           ('NewExternalPort', port),
           ('NewProtocol', protocol)
           ]

    outs = {'NewInternalPort': '',
            'NewInternalClient': '',
            'NewEnabled': '',
            'NewPortMappingDescription': '',
            'NewLeaseDuration': '',
            'errorCode': '714',
            }
    try:
        result = upnp.control(action, ins, outs)
    except UPnPException:
        result = False
    return result, outs

def _addportmapping(upnp, remote, eport, protocol, port, client, description):
    action = 'AddPortMapping'
    ins = [('NewRemoteHost', remote),
           ('NewExternalPort', eport),
           ('NewProtocol', protocol),
           ('NewInternalPort', port),
           ('NewInternalClient', client),
           ('NewEnabled', '1'),
           ('NewPortMappingDescription', description),
           ('NewLeaseDuration', '0'),
           ]
    outs = {}
    result = upnp.control(action, ins, outs)
    return result, outs

def _deleteportmapping(upnp, remote, eport, protocol):
    action = 'DeletePortMapping'
    ins = [('NewRemoteHost', remote),
           ('NewExternalPort', eport),
           ('NewProtocol', protocol),
           ]
    outs = {}
    result = upnp.control(action, ins, outs)
    return result, outs

def _getexternalipaddress(upnp):
    action = 'GetExternalIPAddress'
    ins = []
    outs = {'NewExternalIPAddress': ''}
    result = upnp.control(action, ins, outs)
    return result, outs

def _addmapping(port, eport, client, description, protocol='TCP', timeout=5, remote=''):
    upnp = UPnP()
    # upnp.debug = True
    devices = upnp.discover(timeout=timeout, num=1)
    if devices:
        upnp.descript(devices.pop(0))
    if not upnp.controlurl:
        return ''

    mapped = False

    result, outs = _getspecificportmapping(upnp, remote, eport, protocol)
    if outs['NewInternalPort'] == str(port) \
        and outs['NewInternalClient'] == client \
        and outs['NewPortMappingDescription'] == description:
        mapped = True

    if not mapped:
        try:
            _deleteportmapping(upnp, remote, eport, protocol)
        except:
            pass
        mapped, outs = _addportmapping(upnp, remote, eport, protocol, port, client, description)

    ehost = ''
    if mapped:
        result, outs = _getexternalipaddress(upnp)
        if result:
            ehost = outs.get('NewExternalIPAddress', '')

    outs.clear()
    upnp.close()

    if ehost:
        return '%s:%s' % (ehost, eport)

    return ''

def _deletemapping(eport, protocol='TCP', remote=''):
    upnp = UPnP()
    # upnp.debug = True
    devices = upnp.discover(timeout=timeout, num=1)
    if devices:
        upnp.descript(devices.pop(0))
    if not upnp.controlurl:
        return ''
    return _deleteportmapping(upnp, remote, eport, protocol)


import network
hwaddr = network.gethwaddr()

cache = {}
DEVICEINFO = '/etc/nas/device'
UPNP_EPORT = 'upnpeport'
UPNP_ADDR = 'upnpaddr'
def addmapping(port=80, timeout=5, remote=''):
    ipaddr = network.getipaddr()
    eport, _ = _get_upnp_conf()
    try:
        result = _addmapping(port, eport, ipaddr, hwaddr, timeout=timeout, remote=remote)
    except Exception as error:
        print error
        result = ''
    if result != cache.get(UPNP_ADDR):
        # If there is no result, remove upnpaddr from /etc/nas/device
        values = {UPNP_ADDR: result and result or None}
        if set_name_value(DEVICEINFO, values):
            cache[UPNP_ADDR] = result
    try:
        report_upnp_info()
    except NASError as e: #Ignore possible errors here
        print e

    return result


def _get_upnp_conf():
    conf = {UPNP_EPORT : '', UPNP_ADDR : ''}
    get_name_value(DEVICEINFO, conf)
    eport = conf[UPNP_EPORT]
    if not eport:
        eport = '10080'
    return int(eport), conf[UPNP_ADDR]


def getupnp(**kwargs):
    """Get the external port to use for UPnP mapping"""
    eport, addr = _get_upnp_conf()
    addr = addr.partition(':')[0]
    return {'addr' : addr, 'port' : str(eport)}


def setupnp(port='10080', **kwargs):
    """Set the external port to use for UPnP mapping"""
    if not port.isdigit():
        raise NASError('Port must be a positive number')

    port_number = int(port)
    if port_number < 0 or port_number > 65535:
        raise NASError('Port must be a positive number smaller than 65536')

    oldconf = {UPNP_EPORT : ''}
    get_name_value(DEVICEINFO, oldconf)
    oldeport = oldconf.get(UPNP_EPORT)
    if oldeport:
        try:
            _deletemapping(oldeport)
        except:
            pass #Just try to delete old mapping, ignore any error here

    conf = {UPNP_EPORT : str(port_number)}
    set_name_value(DEVICEINFO, conf)

    result = addmapping(port=80)
    addr = result.partition(':')[0]

    return {'addr' : addr, 'port' : str(port_number)}


__all__ = ['addmapping', 'setupnp', 'getupnp']

# vim: set sta sw=4 et:
